# Parsec

A tiny zero-dependency library for parsing and validating data in TypeScript.

Basically, a very small (<1kb gzipped) alternative to [zod](https://zod.dev)

## Install

Configure your environment to be aware of @endeavorance packages:

```toml
# Bunfig.toml
[install.scopes]
endeavorance = "https://git.astral.camp/api/packages/endeavorance/npm/"
```

Then install the package:

```sh
bun add @endeavorance/parsec
```

## API

Parsec exports two main top-level namespaces: `Parse` and `MakeParserFor`.

- `Parse` contains a set of functions for parsing and validating data.
- `MakeParserFor` contains helpers for creating and composing custom parsers.

## Parse

Functions from the `Parse` namespace return the value as the expected
type if valid. If invalid, these functions throw a `ParseError` with
information about the error.

If a `Parse` function takes multiple arguments, the last argument is
always the value to check in order to allow creating partials easily.

### Primitives

- `Parse.unknown(value: unknown): unknown`
- `Parse.string(value: unknown): string`
- `Parse.number(value: unknown): number`
- `Parse.boolean(value: unknown): boolean`
- `Parse.null(value: unknown): null`
- `Parse.undefined(value: unknown): undefined`
- `Parse.bigint(value: unknown): bigint`
- `Parse.symbol(value: unknown): symbol`

### Arrays

- `Parse.array(value: unknown): unknown[]`
- `Parse.arrayOf<T>(parser: Parser<T>, value: unknown): T[]`

### Object Shapes

To parse a complex object shape, use `Parse.shape`:

```ts
const parsed = Parse.shape("Person", {
  name: Parse.string,
  age: Parse.number,
  isCool: Parse.boolean,
})({
  name: 'John Doe',
  age: 42,
  isCool: true,
});
```

The first parameter is a name for the shape, which is used in error messages.

The second parameter is an object where each key is an expected object property
and each value is the parser for that property.

### Getting Parsed Types

When composing parsers to create more complex types, you can use the `ParsedFrom`
type helper:

```ts
const parsePerson = (value: unknown) => {
  return Parse.shape("Person", {
    name: Parse.string,
    age: Parse.number,
    isCool: Parse.boolean,
  }, value);
}

type Person = ParsedFrom<typeof parsePerson>;
```

### Enums

Parse string enums using `Parse.enum`:

```ts
const WEEKDAYS = ["mon", "tue", "wed", "thu", "fri"] as const;
const parsedWeekday = Parse.enum("Weekday", WEEKDAYS, "mon");
```

### Regex String

For strings that must pass a regex test, use `Parse.regex`:

```ts
const parsed = Parse.regex("AllCaps", /^[A-Z]+$/, "HELLO");
```
## `MakeParserFor`

This object contains utilities for creating custom parsers. These functions
take the same parameters as their `Parse` counterparts except for the value
itself. They return a new function to act as a parser given the provided config.

```ts
const parsePerson = MakeParserFor.shape("Person", {
  name: Parse.string,
  age: Parse.number,
  isCool: Parse.boolean,
});
type Person = ParsedFrom<typeof parsePerson>;
```

## Errors

Parsec exports parsing-specific errors to throw and check for.

- `new ParseError(message: string)`: The base error class for all parsing errors
- `new ParseTypeMismatch(expectedType: string, actualValue: unknown)`: Thrown when the value is not of the expected type



## Branding

Parsec exports a helper type, `Brand<T, B>` which brands type T with the
label provided in B. This is useful for creating custom types that are
primitive types under the hood but have a different name.

```ts
type Integer = Brand<number, "Integer">;
function parseInteger(val: unknown): Integer {
  if (typeof val !== "number" || !Number.isInteger(val)) {
    throw new ParseTypeMismatch("Integer", val);
  }

  return val as Integer;
}
```
